/*
    main.h - Part of sensors, a user-space program for hardware monitoring
    Copyright (C) 2007       Jean Delvare <jdelvare@suse.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    MA 02110-1301 USA.
*/

#ifndef PROG_SENSORS_MAIN_H
#define PROG_SENSORS_MAIN_H

extern int fahrenheit;
extern char degstr[5];

#endif /* PROG_SENSORS_MAIN_H */
